@@ -6,14 +6,12 @@
 //! - [`Type::Union`] is a sum type which resembles a C union.  Each member of the union uses the
 //!   same storage and the size of the union is the size of the largest member.
 //!
-//! [`Type::Contract`] and [`Type::ContractCaller`] are both Sway specific types.
-//!
 //! [`Aggregate`] is an abstract collection of [`Type`]s used for structs, unions and arrays,
 //! though see below for future improvements around splitting arrays into a different construct.
 
 use crate::context::Context;
 
-#[derive(Debug, Clone, Copy, PartialEq)]
+#[derive(Debug, Clone, Copy)]
 pub enum Type {
     Unit,
     Bool,
@@ -23,9 +21,6 @@ pub enum Type {
     Array(Aggregate),
     Union(Aggregate),
     Struct(Aggregate),
-
-    Contract,
-    ContractCaller(AbiInstance),
 }
 
 impl Type {
@@ -58,8 +53,30 @@ impl Type {
                 let agg_content = &context.aggregates[agg.0];
                 format!("{{ {} }}", sep_types_str(agg_content, ", "))
             }
-            Type::Contract => "contract".into(),
-            Type::ContractCaller(_) => "TODO CONTRACT CALLER".into(),
+        }
+    }
+
+    /// Compare a type to this one for equivalence.  We're unable to use `PartialEq` as we need the
+    /// `Context` to compare structs and arrays.
+    pub fn eq(&self, context: &Context, other: &Type) -> bool {
+        match (self, other) {
+            (Type::Unit, Type::Unit) => true,
+            (Type::Bool, Type::Bool) => true,
+            (Type::Uint(l), Type::Uint(r)) => l == r,
+            (Type::B256, Type::B256) => true,
+            (Type::String(l), Type::String(r)) => l == r,
+
+            (Type::Array(l), Type::Array(r)) => l.is_equivalent(context, r),
+            (Type::Struct(l), Type::Struct(r)) => l.is_equivalent(context, r),
+
+            // Unions are special.  We say unions are equivalent to any of their variant types.
+            (Type::Union(l), Type::Union(r)) => l.is_equivalent(context, r),
+            (l, r @ Type::Union(_)) => r.eq(context, l),
+            (Type::Union(l), r) => context.aggregates[l.0]
+                .field_types()
+                .iter()
+                .any(|field_ty| r.eq(context, field_ty)),
+            _ => false,
         }
     }
 }
@@ -75,12 +92,17 @@ impl Type {
 /// But also to keep Type as Copy we need to put the Array meta into another copy type (rather than
 /// recursing with Box<Type>, effectively a different Aggregate.  This could be OK though, still
 /// simpler that what we have here.
+///
+/// NOTE: `Aggregate` derives `Eq` (and `PartialEq`) so that it can also derive `Hash`.  But we must
+/// be careful not to use `==` or `!=` to compare `Aggregate` for equivalency -- i.e., to check
+/// that they represent the same collection of types.  Instead the `is_equivalent()` method is
+/// provided.  XXX Perhaps `Hash` should be impl'd directly without `Eq` if possible?
 
 #[derive(Clone, Copy, Debug, Eq, PartialEq, Hash)]
 pub struct Aggregate(pub generational_arena::Index);
 
 #[doc(hidden)]
-#[derive(Debug, Clone, PartialEq)]
+#[derive(Debug, Clone)]
 pub enum AggregateContent {
     ArrayType(Type, u64),
     FieldTypes(Vec<Type>),
@@ -105,11 +127,17 @@ impl Aggregate {
         )
     }
 
-    /// Get the type of (nested) aggregate fields, if found.
+    /// Tests whether an aggregate has the same sub-types.
+    pub fn is_equivalent(&self, context: &Context, other: &Aggregate) -> bool {
+        context.aggregates[self.0].eq(context, &context.aggregates[other.0])
+    }
+
+    /// Get the type of (nested) aggregate fields, if found.  If an index is into a `Union` then it
+    /// will get the type of the indexed variant.
     pub fn get_field_type(&self, context: &Context, indices: &[u64]) -> Option<Type> {
         indices.iter().fold(Some(Type::Struct(*self)), |ty, idx| {
             ty.and_then(|ty| match ty {
-                Type::Struct(agg) => context.aggregates[agg.0]
+                Type::Struct(agg) | Type::Union(agg) => context.aggregates[agg.0]
                     .field_types()
                     .get(*idx as usize)
                     .cloned(),
@@ -144,30 +172,20 @@ impl AggregateContent {
             AggregateContent::ArrayType(ty, cnt) => (ty, cnt),
         }
     }
-}
 
-/// A Sway specific data structure for associating an ABI with an address.
-#[derive(Clone, Copy, Debug, Eq, PartialEq, Hash)]
-pub struct AbiInstance(pub generational_arena::Index);
+    /// Tests whether an aggregate has the same sub-types.
+    pub fn eq(&self, context: &Context, other: &AggregateContent) -> bool {
+        match (self, other) {
+            (AggregateContent::FieldTypes(l_tys), AggregateContent::FieldTypes(r_tys)) => l_tys
+                .iter()
+                .zip(r_tys.iter())
+                .all(|(l, r)| l.eq(context, r)),
+            (
+                AggregateContent::ArrayType(l_ty, l_cnt),
+                AggregateContent::ArrayType(r_ty, r_cnt),
+            ) => l_cnt == r_cnt && l_ty.eq(context, r_ty),
 
-#[doc(hidden)]
-#[derive(Debug, Clone, PartialEq)]
-pub struct AbiInstanceContent {
-    pub name: Vec<String>,
-    pub address: String,
-}
-
-impl AbiInstance {
-    pub fn new(
-        context: &mut Context,
-        mut name_prefixes: Vec<String>,
-        name_suffix: String,
-        address: String,
-    ) -> Self {
-        name_prefixes.push(name_suffix);
-        AbiInstance(context.abi_instances.insert(AbiInstanceContent {
-            name: name_prefixes,
-            address,
-        }))
+            _ => false,
+        }
     }
 }
